// <<hpp_insert gen/CellRenderer.js>>

class OmScalarScalarCell extends ScalarBase {
    constructor(color, id) {
        super(color, id);
    }
}

class OmScalarVectorCell extends VectorBase {
    constructor(color, id) {
        super(color, id);
    }
}

class OmVectorScalarCell extends VectorBase {
    constructor(color, id) {
        super(color, id);
    }
}

class OmVectorVectorCell extends VectorBase {
    constructor(color, id) {
        super(color, id);
    }
}

class OmScalarGroupCell extends GroupBase {
    constructor(color, id) {
        super(color, id);
    }
}

class OmGroupScalarCell extends GroupBase {
    constructor(color, id) {
        super(color, id);
    }
}

class OmVectorGroupCell extends GroupBase {
    constructor(color, id) {
        super(color, id);
    }
}

class OmGroupVectorCell extends GroupBase {
    constructor(color, id) {
        super(color, id);
    }
}

class OmGroupGroupCell extends GroupBase {
    constructor(color, id) {
        super(color, id);
    }
}
